package polyglot.ext.atomjava.ast;

import polyglot.ext.jl.ast.ArrayAccessAssign_c;
import polyglot.util.Position;
import polyglot.ast.*;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.visit.*;

import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.ext.atomjava.util.AJS;
import polyglot.ext.atomjava.frontend.*;

import java.util.List;

public class AtomJavaArrayAccessAssign_c 
  extends ArrayAccessAssign_c
  implements AtomJavaBuild {

  public AtomJavaArrayAccessAssign_c(Position pos, ArrayAccess left, Operator op, Expr right) {
    super(pos, left, op, right);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    AJOptions options = (AJOptions) ab.job().extensionInfo().getOptions();

    if(options.optimize &&
       ((ArrayAccess)left).array() instanceof Field) {

      Field f = (Field) ((ArrayAccess)left).array();

      if(f.isTargetImplicit() &&
         !NotInAtomicOptimizer.fieldArraySubscriptAssignedInAtomic(f.fieldInstance()) &&
         !NotInAtomicOptimizer.fieldArraySubscriptAccessedInAtomic(f.fieldInstance()) &&
         // not aliased (via a field read in either atomic or non-atomic)
         !NotInAtomicOptimizer.fieldArrayRead(f.fieldInstance()) &&
         // not written (which includes in atomic) [ technically doesn't matter ]
         !NotInAtomicOptimizer.fieldArrayWritten(f.fieldInstance())) {
        System.out.println("OPTIMIZING ARRAY ASSIGN " + f.fieldInstance());
        return buildWeak(parent, ab);
      }
    }

    if(options.weak && !ab.context().inAtomicContext())
      return buildWeak(parent, ab);

    return buildStrong(parent, ab);
  }

  public Node buildWeak(Node parent, AtomJavaBuilder ab) {
    return this;
  }

  public Node buildStrong(Node parent, AtomJavaBuilder ab) {
    AtomJavaContext     ajc = ab.context();
    AtomJavaNodeFactory nf  = ab.nodeFactory();
    AtomJavaTypeSystem  ts  = ab.typeSystem();
        
    Expr array = ((ArrayAccess)left).array();
    Expr index = ((ArrayAccess)left).index();

    Type base = array.type().toArray().base();
    List call_args;

    if((base.isChar() || 
       base.isByte() || 
       base.isShort()) &&
       !ts.equals(base, right.type()) &&
       !ts.isImplicitCastValid(right.type(), base)) {
      call_args = 
        AJS.List(array, index,
                 nf.Cast(ajc, nf.CanonicalTypeNode(ajc, base), right));
    } else call_args = AJS.List(array,index,right);
    
    if(op == ASSIGN)
      return nf.Call(ajc, nf.CanonicalTypeNode(ajc, ts.AArray()), 
                     ajc.arrayElementSetterName(base), call_args);

    return nf.Call(ajc, null, call_args,
                   ajc.makeArrayOperatorMethod(array.type(),op,ab));
  }
}

