package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.frontend.*;
import polyglot.ext.atomjava.visit.*;

import polyglot.ext.jl.ast.ArrayAccess_c;
import polyglot.util.CodeWriter;
import polyglot.util.Position;
import polyglot.ast.*;
import polyglot.visit.*;

import java.util.List;
import java.util.LinkedList;


public class AtomJavaArrayAccess_c 
  extends ArrayAccess_c 
  implements AtomJavaBuild {

  public AtomJavaArrayAccess_c(Position pos, Expr array, Expr index) {
    super(pos, array, index);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    AJOptions options = (AJOptions) ab.job().extensionInfo().getOptions();

    if(options.optimize && array instanceof Field) {
      
      Field f = (Field) array;

      if(f.isTargetImplicit() &&
         !NotInAtomicOptimizer.fieldArraySubscriptAssignedInAtomic(f.fieldInstance()) &&
         !NotInAtomicOptimizer.fieldArraySubscriptAccessedInAtomic(f.fieldInstance()) &&
         // not aliased (via a field read in either atomic or non-atomic)
         !NotInAtomicOptimizer.fieldArrayRead(f.fieldInstance()) &&
         // not written (which includes in atomic) [ technically doesn't matter ]
         !NotInAtomicOptimizer.fieldArrayWritten(f.fieldInstance())) {
        System.out.println("OPTIMIZING ARRAY ACCESS " + f.fieldInstance());
        return buildWeak(parent, ab);
      }
    }
       
    if(options.weak && !ab.context().inAtomicContext())
      return buildWeak(parent, ab);

    return buildStrong(parent, ab);
  }

  public Node buildWeak(Node parent, AtomJavaBuilder ab) {
    return this;
  }

  public Node buildStrong(Node parent, AtomJavaBuilder ab) { 
    AtomJavaContext ajc = (AtomJavaContext) ab.context();
    AtomJavaNodeFactory nf = (AtomJavaNodeFactory) ab.nodeFactory();
    AtomJavaTypeSystem ts = (AtomJavaTypeSystem) ab.typeSystem();

    List args = new LinkedList();
    args.add(array);
    args.add(index);
    
    return nf.Call(ajc, nf.CanonicalTypeNode(ajc, ts.AArray()), 
                   ajc.arrayElementGetterName(array.type().toArray().base()), 
                   args);
  }
}
