package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.visit.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.jl.ast.ConstructorCall_c;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.ast.*;
import java.util.*;

/** ConstructorCall's in an atomic context need to be converted to call the 
 *  version of the contructor whose body had been prepared for being
 *  inside an atomic context (the constructor which has been marked by an
 *  extra arguments: acons). However, there is the potential that a super
 *  call might not be valid. For example, given that someone has extended
 *  a class that has not been prepared for atomic calls (or can not be:
 *  an untouchable) then the only alternative is to leave the super call
 *  and assume that we fail at the call to new. The concern of failing at
 *  the super or this call site is that if we don't provide a super call
 *  then either it will fail at compile time or the default constructor will
 *  get called instead and certain side effects might occur because part of
 *  the object will get created.
 */

public class AtomJavaConstructorCall_c 
  extends ConstructorCall_c 
  implements AtomJavaBuild
{
  public AtomJavaConstructorCall_c(Position pos, Kind kind, Expr qualifier, List arguments) {
    super(pos,kind,qualifier,arguments);
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    if(ab.context().inAtomicContext())
      return buildAtomic(ab);
    return buildNotAtomic(ab);
  }

  public Node buildAtomic(AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    List args = new LinkedList(arguments);
    Field f = nf.Field(c, ts.AConstructorFormal(), 
                       c.atomicConstructorFieldName());
    args.add(0, f);
    /* try and find the constructor, super or this */
    try {
      ConstructorInstance ci = 
        ts.findConstructor(constructorInstance().container().toClass(), 
                           AJS.Expr_to_Type(args), c.currentClass());
      /* found the constructor */
      ConstructorCall cc = (ConstructorCall) arguments(args);
      return cc.constructorInstance(ci);
    } catch (SemanticException e) {}

    /* JAVAC COMPILER HACK */ return this;
  }

  public Node buildNotAtomic(AtomJavaBuilder ab) {
    return this;
  }

}
