package polyglot.ext.atomjava.ast;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;
import polyglot.ext.jl.ast.*;
import java.util.*;

/**
 * An AST node for an <code>AtomicBlock</code>. 
 */
public interface AtomicBlock extends CompoundStmt 
{
  /**
   * AtomicBlockEntry
   * Provides a plug into the data flow analysis such that we are
   * gauranteed that we can do any important work BEFORE entering the block
   * @see polyglot.ext.atomjava.visit.AJInitChecker   
   */
  class AtomicBlockEntry extends Term_c {
    /* a list of <code>LocalInstance</code> that are initialized.
     *  @see polyglot.ext.atomjava.visit.AJInitChecker#check
     */
    List uninitialized;

    public AtomicBlockEntry() {
      super(Position.COMPILER_GENERATED);
    }
    
    public void uninitializedDecls(List list) {
      uninitialized = list;
    }
    
    public List uninitializedDecls() {
      return uninitialized;
    }
    
    public Term entry() {
      return this;
    }
    
    public List acceptCFG(CFGBuilder v, List succs) {
      return succs;
    }
  }


}
