package polyglot.ext.atomjava.ast;

import polyglot.ext.atomjava.types.AtomJavaContext;
import polyglot.ext.atomjava.types.AtomJavaTypeSystem;
import polyglot.ext.atomjava.types.AtomJavaRuntimeType;

import polyglot.ext.jl.ast.*;
import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;
import java.util.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.atomjava.visit.*;
import polyglot.types.LocalInstance;

/**
 * An AST node for a <code>AtomicBlock</code>. 
 */
public class AtomicBlock_c 
  extends AbstractBlock_c /* crucial for scope of vars translation introduces */
  implements AtomicBlock, AtomJavaBuild 
{

  /* the Term upon which we enter this block */
  protected AtomicBlockEntry entry;

  public AtomicBlock_c(Position pos, List statements) {
    super(pos,statements);
    this.entry = new AtomicBlockEntry();
  }

  // DAN: see the hack below where we look in the outer context.
  //      This is probably not the right way to do this.
  public Context enterScope(Context c) {
    AtomJavaContext ajc = (AtomJavaContext) super.enterScope(c);
    return ((AtomJavaContext)ajc.pushAtomicBlock()).pushAtomic();
  }

  public NodeVisitor buildEnter(AtomJavaBuilder ab) {
    return ab;
  }

  public Node build(Node parent, AtomJavaBuilder ab) {
    if(((AtomJavaContext)ab.context().pop()).inAtomicContext())
      return buildAtomic(ab);
    return buildNotAtomic(ab);
  }

  public Node buildAtomic(AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaNodeFactory nf = ab.nodeFactory();
    return nf.Block(c,statements);
  }

  public Node buildNotAtomic(AtomJavaBuilder ab) {
    AtomJavaContext     c  = ab.context();
    AtomJavaTypeSystem  ts = ab.typeSystem();
    AtomJavaNodeFactory nf = ab.nodeFactory();

    /* the block that holds all things atomic */
    Block b = nf.Block(c, Collections.EMPTY_LIST);

    List local_assigns = new LinkedList();

    for(Iterator locals = c.getLocalVars().iterator(); locals.hasNext();) {
      LocalInstance li = (LocalInstance)locals.next();

      // we assume all locals are initialized

      // if(entry.uninitializedDecls().contains(li))
      //   continue;

      if(li.flags().isFinal())
	continue;
      
      LocalDecl ld = nf.freshLocalDecl(c, Flags.NONE, li.name(), 
                                       li.type(), nf.Local(c, li));      
      // int __aj_i_ = i;
      b = b.append(ld);
      
      Assign local_assign = nf.Assign(c, nf.Local(c, li), nf.Local(c, ld));
      
      local_assigns.add(nf.Eval(c, local_assign));
    }
    
    LocalDecl done = 
      nf.freshLocalDecl(c, Flags.NONE, "done", ts.Boolean(), 
                        nf.BooleanLit(c, false));

    // boolean done = false;     
    b = b.append(done);
    
    Call start_atomic = nf.Call(c, nf.currentThread(c), 
                                "start_atomic", Collections.EMPTY_LIST);

    //AThread.currentThread(c).start_atomic();
    b = b.append(nf.Eval(c, start_atomic));
    
    Assign done_true = 
      nf.Assign(c, nf.Local(c, done), nf.BooleanLit(c, true));
    
    Assign done_doomed_false = 
      nf.Assign(c, nf.Local(c, done), 
                nf.Assign(c, nf.threadLocalField(c, "doomed"), 
                          nf.BooleanLit(c, false)));
    
    Call rollback_and_sleep =
      nf.Call(c, nf.currentThread(c), "rollback_and_sleep",
              Collections.EMPTY_LIST);
    
    Block if_doomed_consequent = nf.Block(c, local_assigns);
    if_doomed_consequent = 
      if_doomed_consequent.prepend(nf.Eval(c, done_doomed_false));
    if_doomed_consequent = 
      if_doomed_consequent.append(nf.Eval(c, rollback_and_sleep));

    If if_doomed = 
      nf.If(c, nf.threadLocalField(c, "doomed"),
            if_doomed_consequent, null);
    
    Call end_atomic = nf.Call(c, nf.currentThread(c), "end_atomic",
                              Collections.EMPTY_LIST);
    
    If if_done = nf.If(c, nf.Local(c, done), 
                       nf.Eval(c, end_atomic),
                       nf.Branch(c, Branch.CONTINUE, null));
    
    Formal f = nf.Formal(c, Flags.NONE, 
                         nf.CanonicalTypeNode(c, ts.AtomJavaRollBack()), "rb");
   
    Block catch_block = nf.Block(c, local_assigns);
    catch_block = catch_block.prepend(nf.Eval(c, done_doomed_false));
    catch_block = catch_block.append(nf.Eval(c, rollback_and_sleep));
    Catch ctch = nf.Catch(c, f, catch_block);

    Block finally_block = nf.Block(c, AJS.List(if_doomed, if_done));

    Try t = nf.Try(c, nf.Block(c,statements), AJS.List(ctch), finally_block);
    
//     While w = nf.While(c, nf.Unary(c, Unary.NOT, nf.Local(c, done)),
//                        nf.Block(c, AJS.List(nf.Eval(c, done_true), t)));

    Do d = nf.Do(c, nf.Block(c, AJS.List(nf.Eval(c, done_true), t)),
		 nf.Unary(c, Unary.NOT, nf.Local(c, done)));

    b = b.append(d);
    
    if(!reachable) {
      New n = 
        nf.New(c, null, 
               nf.CanonicalTypeNode(c, ts.AtomJavaUnreachableCodeError()), 
               Collections.EMPTY_LIST, null);
       b = b.append(nf.Throw(c, n));
    }
    return b;
  }

  /** Visit the children of the statement. 
      @return <code>this</code> or a copy if our children have changed */
  public Node visitChildren(NodeVisitor v) {
    visitChild(entry, v); // necessary?
    List ss = visitList(this.statements, v);
    if(ss != statements) {
      AtomicBlock_c ab = (AtomicBlock_c) copy();
      ab.statements = TypedList.copyAndCheck(ss, Stmt.class, true);
      return ab;
    } 
    return this;
  }

  public String toString() {
    return "atomic { ... } ";
  }
  
  public Term entry() {
    return entry;
  }
  
  public List acceptCFG(CFGBuilder v, List succs) {
    v.visitCFGList(AJS.cons(entry,new LinkedList(statements)),this);
    return succs;
  }
}
