package polyglot.ext.atomjava.main;

import polyglot.main.Options;
import polyglot.frontend.ExtensionInfo;
import polyglot.main.UsageError;
import polyglot.main.Main;

import java.util.*;
import java.io.*;

public class AJOptions extends Options {
  public AJOptions(ExtensionInfo extension) {
    super(extension);
  }

  public boolean weak = false;
  public boolean optimize = false;
  private boolean output_directory_specified = false;
  public String polyglot;

  public void setDefaultValues() {
    super.setDefaultValues();
    polyglot = System.getenv("POLYGLOT");
    System.out.println("polyglot: " + polyglot);
  }

  public void parseCommandLine(String args[], Set source) throws UsageError {

    super.parseCommandLine(args, source);
    
    if(optimize) {

      if(weak)
	throw new UsageError("'-optimize' and '-weak' are not compatible.");

      if(polyglot == null)
	throw new UsageError("'-optimize' expects '-polyglot dir'.");

      // this is commented out because we should just force the programmer
      // to do this instead
//       // add 'src' to the path such that we can find atomized j2se source 
//       // (not wrappers because source loader only finds '.ajava')
//       String path = polyglot + File.separator + "src";
//       File f = new File(path);
//       if (f != null && !source_path.contains(f))
// 	source_path.add(f);

      if(!output_directory_specified)
	output_directory = new File(System.getProperty("user.dir") 
				    + File.separator + "optimized");
    }
  }

  protected int parseCommand(String args[], int index, Set source) 
    throws UsageError, Main.TerminationException {

    int i = index;

    if(args[i].equals("-weak")) {
      i++;
      weak = true;
    }
    else if(args[i].equals("-optimize")) {
      i++;
      optimize = true;
    }
    else if(args[i].equals("-polyglot")) {
      i++;
      polyglot = args[i];
      i++;
    }
    else if (args[i].equals("-d")) {
      i++;
      output_directory = new File(args[i]);
      i++;      
      output_directory_specified = true;
    }
    
    if(i >= args.length) return i;
    else return super.parseCommand(args, i, source);
  }

  public String constructFullClasspath() {
    if(!optimize)
      return super.constructFullClasspath();

    String bootclasspath = System.getProperty("sun.boot.class.path");
    if (bootclasspath == null) {
      bootclasspath = System.getProperty("java.home") +
        File.separator + "jre" +
        File.separator + "lib" +
        File.separator + "rt.jar";
    }
    bootclasspath += File.pathSeparator + polyglot + File.separator + 
      "classes";
    bootclasspath += File.pathSeparator + polyglot + File.separator + 
      "lib" + File.separator + "polyglot.jar";
    bootclasspath += File.pathSeparator + polyglot + File.separator +
      "lib" + File.separator + "java_cup.jar";
    bootclasspath += File.pathSeparator + polyglot + File.separator + 
      "lib" + File.separator + "atomjava.jar";
    return bootclasspath;
  }

  public void usage(PrintStream out) {
    super.usage(out);
    out.println("---- AtomJava Options ----");
    usageForFlag(out, "-weak", "compile for weak atomicity");
    usageForFlag(out, "-optimize", "whole-program optimization");
    usageForFlag(out, "         ", "-classpath is not supported in presence of optimization");
    usageForFlag(out, "         ", "-polyglot dir is expected for optimization");
  }  
}
