package polyglot.ext.atomjava.runtime;

// every AObject is an AtomJava lock and has a current-lock.
// Policy: objects start being locked by self
public class AObject extends Object {
  // public AObject currentLock;   // my lock
  public AThread currentHolder; // as a lock, who holds me
  
  public AObject() {
    //    currentLock   = this;
    currentHolder = AThread.currentThread(); 
  }
  public AObject(AThread holder) {
    //    currentLock = this;
    currentHolder = holder;
  }
  public AObject(AConstructorFormal acons) {
    this();
  }
  public final void acq_mylk() { // acq my current lock
    acq_mylk(AThread.currentThread());
  }
  public final void acq_mylk(AThread me) { // acq my current lock
    //    AObject lk = currentLock;
    if(/*currentLock.*/currentHolder == me) {
      return;
    }
    me.acq_lk(this/*lk*/);
    /*    if(lk != currentLock) { // needed only when locks change dynamically
      me.rel_lk(lk); // unnecessary but a good idea?
      me.check_for_release(); // just to be careful because in a loop
      acq_mylk(me); // very unlikely tail-call
      }*/
  }

    /***********************************************************************/
    /*             _   _   ___ ___ ___ ___    ___ ___ _    _ ___           */
    /*            | | |_)   |  |_  |    |      |  |_|  \  /  |_|           */
    /*            |_| |_) (_|  |__ |__  | o (__|  | |   \/   | |           */
    /*  override and provide atomic versions of methods in Object          */
    /***********************************************************************/

  /**************************************************************************/
  /*    atomic versions of methods, super calls b/c dynamic dispatch        */
  /**************************************************************************/
  public int     __aj_atomic_hashCode()         { return super.hashCode();  }
  public boolean __aj_atomic_equals(Object obj) { return super.equals(obj); }
  public Class   __aj_atomic_getClass()         { return super.getClass();  }
  public String  __aj_atomic_toString()         { return super.toString();  }
  protected void __aj_atomic_finalize() throws Throwable { super.finalize(); }

  /**************************************************************************/
  /*    atomic versions of methods, throws AtomJavaNativeInAtomicError      */
  /**************************************************************************/
  public void __aj_atomic_notify() {
    throw new AtomJavaNativeInAtomicError("notify");
  }
  public void __aj_atomic_notifyAll() {
    throw new AtomJavaNativeInAtomicError("notifyAll");
  }

  /**************************************************************************/
  /*                overriden methods which require locks                   */
  /**************************************************************************/
  protected Object clone() throws CloneNotSupportedException {
    acq_mylk();
    return super.clone();
  }
  protected Object __aj_atomic_clone() throws CloneNotSupportedException {
    acq_mylk();
    return super.clone();
  }

  /**************************************************************************/
  /*       final methods that are mangled by the atomjava compiler          */
  /**************************************************************************/
  public void __aj_mangle_wait() throws InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.wait(); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public void __aj_mangle_wait(long timeout, int nanos) throws InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.wait(timeout, nanos); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public void __aj_mangle_wait(long timeout) throws InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.wait(timeout); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }

  // atomic versions
  public void __aj_atomic___aj_mangle_wait() 
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
  public void __aj_atomic___aj_mangle_wait(long timeout, int nanos) 
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
  public void __aj_atomic___aj_mangle_wait(long timeout) 
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
}
