package polyglot.ext.atomjava.runtime;
import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

///////////////////////////////////////////////////////////////////////////////
// TO DO: Because AThread is not final, all fields and methods need their    //
//  names mangled (prepended with __aj_mangle or see visit/CallMangler.java  //
// Only those methods which could ever be called (public) those which are    //
// known about (from Thread) or those which we explictely insert into code   //
// and want to be considered as acting differently whether its in            //
// atomic or not                                                             //
///////////////////////////////////////////////////////////////////////////////

public class AThread extends Thread {

  /**********************************************************************/ 
  /*   logging and rollback                                             */
  /**********************************************************************/ 

  private final static Object[] resize_object_arr(Object[] arr) {
    Object[] tmp = new Object[arr.length << 1];
    System.arraycopy(arr, 0, tmp, 0, arr.length);
    return tmp;
  }
  private final static int[] resize_int_arr(int[] arr) {
    int[] tmp = new int[arr.length << 1];
    System.arraycopy(arr, 0, tmp, 0, arr.length);
    return tmp;
  }

  /******** Logging Object's *********/
  int object_index;
  Object[] object_containers;
  UndoObject[] object_undoers;
  Object[] object_originals;
  int object_compacted_index;

  private final void cull_compact_resize_objects_log() {
    /* remove duplicates */
    for(int i = 0; i <= object_index; i++) {
      if(object_undoers[i] == null) continue;
      for(int n=Math.max(object_compacted_index,i+1); n<=object_index; n++) {
        if(object_undoers[n] == null) continue;
        if((object_undoers[i] == object_undoers[n]) && 
           (object_containers[i] == object_containers[n])) {
          object_undoers[n] = null;
          object_containers[n] = null;
          object_originals[n] = null;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = object_index;
      object_index = object_compacted_index;
      for(; object_index <= pointer; object_index++) {
        if(object_undoers[object_index] != null) 
          continue;
        
        while(object_undoers[pointer] == null && pointer > object_index) 
          pointer--;
        
        if(object_index >= pointer)
          break compact;
        
        object_containers[object_index] = object_containers[pointer];
        object_undoers[object_index] = object_undoers[pointer];
        object_originals[object_index] = object_originals[pointer];
        object_containers[pointer] 
          = object_originals[pointer] 
          = object_undoers[pointer] = null;
      }
      break compact;
    }
    object_compacted_index = object_index--;
    /* resize */
    if((object_index+1) <= (object_containers.length/2))
      return;
    
    object_containers = resize_object_arr(object_containers);
    UndoObject[] undoers = new UndoObject[object_undoers.length << 1];
    System.arraycopy(object_undoers, 0, undoers, 0, object_undoers.length);
    object_undoers = undoers;
    object_originals = resize_object_arr(object_originals);    
  }

  final private void rollback_objects_log() {
    int i = object_index;
    for(;i >= 0; i--)
      object_undoers[i].undo(object_containers[i],object_originals[i]);
    object_index = -1;
    object_compacted_index = 0;
  }

  final public void log(Object obj, UndoObject u, Object orig) {
    if(object_index == object_containers.length - 1)
      cull_compact_resize_objects_log();

    object_containers[++object_index] = obj;
    object_undoers[object_index] = u;
    object_originals[object_index] = orig;
  }

  /******** Logging Integer's *********/
  int integer_index;
  Object[] integer_containers;
  UndoInteger[] integer_undoers;
  long[] integer_originals; // boolean, byte, char, short, int, long
  int integer_compacted_index;

  private final void cull_compact_resize_integers_log() {
    /* remove duplicates */
    for(int i = 0; i <= integer_index; i++) {
      if(integer_undoers[i] == null) continue;
      for(int n=Math.max(integer_compacted_index,i+1); n<=integer_index; n++) {
        if(integer_undoers[n] == null) continue;
        if((integer_undoers[i] == integer_undoers[n]) && 
           (integer_containers[i] == integer_containers[n])) {
          integer_undoers[n] = null;
          integer_containers[n] = null;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = integer_index;
      integer_index = integer_compacted_index;
      for(; integer_index <= pointer; integer_index++) {
        if(integer_undoers[integer_index] != null) 
          continue;
        
        while(integer_undoers[pointer] == null && pointer > integer_index) 
          pointer--;
        
        if(integer_index >= pointer)
          break compact;
        
        integer_containers[integer_index] = integer_containers[pointer];
        integer_undoers[integer_index] = integer_undoers[pointer];
        integer_originals[integer_index] = integer_originals[pointer];
        integer_containers[pointer] = integer_undoers[pointer] = null;
      }
      break compact;
    }
    integer_compacted_index = integer_index--;
    /* resize */
    if((integer_index+1) <= (integer_containers.length/2))
      return;
    
    integer_containers = resize_object_arr(integer_containers);
    UndoInteger[] undoers = new UndoInteger[integer_undoers.length << 1];
    System.arraycopy(integer_undoers, 0, undoers, 0, integer_undoers.length);
    integer_undoers = undoers;
    long[] originals = new long[integer_originals.length << 1];
    System.arraycopy(integer_originals, 0, originals, 0, integer_originals.length);
    integer_originals = originals;
  }

  private final void rollback_integers_log() {
    int i = integer_index;
    for(;i >= 0; i--)
      integer_undoers[i].undo(integer_containers[i],integer_originals[i]);
    integer_index = -1;
    integer_compacted_index = 0;
  }

  public final void log(Object obj, UndoInteger u, long orig) {
    if(integer_index == integer_containers.length - 1)
      cull_compact_resize_integers_log();

    integer_containers[++integer_index] = obj;
    integer_undoers[integer_index] = u;
    integer_originals[integer_index] = orig;
  }
  
  /******** Logging Real's *********/
  int real_index;
  Object[] real_containers; 
  UndoReal[] real_undoers;
  double[] real_originals; // float, double
  int real_compacted_index;

  private final void cull_compact_resize_reals_log() {
    /* remove duplicates */
    for(int i = 0; i <= real_index; i++) {
      if(real_undoers[i] == null) continue;
      for(int n=Math.max(real_compacted_index,i+1); n<=real_index; n++) {
        if(real_undoers[n] == null) continue;
        if((real_undoers[i] == real_undoers[n]) && 
           (real_containers[i] == real_containers[n])) {
          real_undoers[n] = null;
          real_containers[n] = null;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = real_index;
      real_index = real_compacted_index;
      for(; real_index <= pointer; real_index++) {
        if(real_undoers[real_index] != null) 
          continue;
        
        while(real_undoers[pointer] == null && pointer > real_index) 
          pointer--;
        
        if(real_index >= pointer)
          break compact;
        
        real_containers[real_index] = real_containers[pointer];
        real_undoers[real_index] = real_undoers[pointer];
        real_originals[real_index] = real_originals[pointer];
        real_containers[pointer] = real_undoers[pointer] = null;
      }
      break compact;
    }
    real_compacted_index = real_index--;
    /* resize */
    if((real_index+1) <= (real_containers.length/2))
      return;

    real_containers = resize_object_arr(real_containers);
    UndoReal[] undoers = new UndoReal[real_undoers.length << 1];
    System.arraycopy(real_undoers, 0, undoers, 0, real_undoers.length);
    real_undoers = undoers;
    double[] originals = new double[real_originals.length << 1];
    System.arraycopy(real_originals, 0, originals, 0, real_originals.length);
    real_originals = originals;    
  }

  private final void rollback_reals_log() {
    int i = real_index;
    for(;i >= 0; i--)
      real_undoers[i].undo(real_containers[i],real_originals[i]);
    real_index = -1;
    real_compacted_index = 0;
  }

  public final void log(Object obj, UndoReal u, double orig) {
    if(real_index == real_containers.length - 1)
      cull_compact_resize_reals_log();

    real_containers[++real_index] = obj;
    real_undoers[real_index] = u;
    real_originals[real_index] = orig;
  }

  /******** Logging Object Array's (subscripts) *********/
  int object_arrays_index;
  Object[][] object_arrays_arrays;
  int[] object_arrays_subscript;
  Object[] object_arrays_originals;
  int object_arrays_compacted_index;

  private final void cull_compact_resize_object_arrays_log() {
    /* remove duplicates */
    for(int i = 0; i <= object_arrays_index; i++) {
      if(object_arrays_arrays[i] == null) continue;
      for(int n=Math.max(object_arrays_compacted_index,i+1); 
          n<=object_arrays_index; n++) {
        if(object_arrays_arrays[n] == null) continue;
        if((object_arrays_arrays[i] == object_arrays_arrays[n]) && 
           (object_arrays_subscript[i] == object_arrays_subscript[n])) {
          object_arrays_arrays[n] = null;
          object_arrays_subscript[n] = -1;
          object_arrays_originals[n] = null;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = object_arrays_index;
      object_arrays_index = object_arrays_compacted_index;
      for(; object_arrays_index <= pointer; object_arrays_index++) {
        if(object_arrays_arrays[object_arrays_index] != null) 
          continue;
        
        while(object_arrays_arrays[pointer] == null && 
              pointer > object_arrays_index) 
          pointer--;
        
        if(object_arrays_index >= pointer)
          break compact;
        
        object_arrays_arrays[object_arrays_index] = 
          object_arrays_arrays[pointer];
        object_arrays_subscript[object_arrays_index] = 
          object_arrays_subscript[pointer];
        object_arrays_originals[object_arrays_index] = 
          object_arrays_originals[pointer];
        object_arrays_originals[pointer] = 
          object_arrays_arrays[pointer] = null;
        object_arrays_subscript[pointer] = -1;
      }
      break compact;
    }
    object_arrays_compacted_index = object_arrays_index--;
    /* resize */
    if((object_arrays_index+1) <= (object_arrays_arrays.length/2))
      return;

    Object[][] arrays = new Object[object_arrays_arrays.length << 1][];
    System.arraycopy(object_arrays_arrays, 0, arrays, 0, object_arrays_arrays.length);
    object_arrays_arrays = arrays;
    object_arrays_subscript = resize_int_arr(object_arrays_subscript);
    object_arrays_originals = resize_object_arr(object_arrays_originals);
  }

  private final void rollback_object_arrays_log() {
    int i = object_arrays_index;
    for(;i >= 0; i--)
      object_arrays_arrays[i][object_arrays_subscript[i]] = 
        object_arrays_originals[i];
    object_arrays_index = -1;
    object_arrays_compacted_index = 0;
  }

  public final void log(Object[] array, int subscript, Object orig) {
    if(object_arrays_index == object_arrays_arrays.length - 1)
      cull_compact_resize_object_arrays_log();

    object_arrays_arrays[++object_arrays_index] = array;
    object_arrays_subscript[object_arrays_index] = subscript;
    object_arrays_originals[object_arrays_index] = orig;
  }

  /******** Logging Integer Array's (subscripts) *********/
  int integer_arrays_index;
  Object[] integer_arrays_arrays;
  int[] integer_arrays_subscript;
  long[] integer_arrays_originals; // boolean, byte, char, short, int, long
  int integer_arrays_compacted_index;

  private final void cull_compact_resize_integer_arrays_log() {
    /* remove duplicates */
    for(int i = 0; i <= integer_arrays_index; i++) {
      if(integer_arrays_arrays[i] == null) continue;
      for(int n=Math.max(integer_arrays_compacted_index,i+1); 
          n<=integer_arrays_index; n++) {
        if(integer_arrays_arrays[n] == null) continue;
        if((integer_arrays_arrays[i] == integer_arrays_arrays[n]) && 
           (integer_arrays_subscript[i] == integer_arrays_subscript[n])) {
          integer_arrays_arrays[n] = null;
          integer_arrays_subscript[n] = -1;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = integer_arrays_index;
      integer_arrays_index = integer_arrays_compacted_index;
      for(; integer_arrays_index <= pointer; integer_arrays_index++) {
        if(integer_arrays_arrays[integer_arrays_index] != null) 
          continue;
        
        while(integer_arrays_arrays[pointer] == null && 
              pointer > integer_arrays_index) 
          pointer--;
        
        if(integer_arrays_index >= pointer)
          break compact;
        
        integer_arrays_arrays[integer_arrays_index] = 
          integer_arrays_arrays[pointer];
        integer_arrays_subscript[integer_arrays_index] = 
          integer_arrays_subscript[pointer];
        integer_arrays_originals[integer_arrays_index] = 
          integer_arrays_originals[pointer];
        integer_arrays_arrays[pointer] = null;
        integer_arrays_subscript[pointer] = -1;
      }
      break compact;
    }
    integer_arrays_compacted_index = integer_arrays_index--;
    /* resize */
    if((integer_arrays_index+1) <= (integer_arrays_arrays.length/2))
      return;

    integer_arrays_arrays = resize_object_arr(integer_arrays_arrays);
    integer_arrays_subscript = resize_int_arr(integer_arrays_subscript);
    long[] originals = new long[integer_arrays_originals.length << 1];
    System.arraycopy(integer_arrays_originals, 0, originals, 0, integer_arrays_originals.length);
    integer_arrays_originals = originals;   
  }

  private final void rollback_integer_arrays_log() {
    int i = integer_arrays_index;
    for(;i >= 0; i--) {
      Object arr  = integer_arrays_arrays[i];
      int    sub  = integer_arrays_subscript[i];
      long   orig = integer_arrays_originals[i];
      if(arr instanceof boolean[]) ((boolean[])arr)[sub] = (orig==1?true:false);
      else if(arr instanceof byte[])  ((byte[])arr)[sub]  = (byte)orig;
      else if(arr instanceof char[])  ((char[])arr)[sub]  = (char)orig;
      else if(arr instanceof short[]) ((short[])arr)[sub] = (short)orig;
      else if(arr instanceof int[])   ((int[])arr)[sub]   = (int)orig;
      else                            ((long[])arr)[sub]  = orig;
    }
    integer_arrays_index = -1;
    integer_arrays_compacted_index = 0;
  }

  public final void log(Object array, int subscript, long orig) {
    if(integer_arrays_index == integer_arrays_arrays.length - 1)
      cull_compact_resize_integer_arrays_log();

    integer_arrays_arrays[++integer_arrays_index] = array;
    integer_arrays_subscript[integer_arrays_index] = subscript;
    integer_arrays_originals[integer_arrays_index] = orig;
  }

  /******** Logging Real Array's (subscripts) *********/
  int real_arrays_index;
  Object[] real_arrays_arrays;
  int[] real_arrays_subscript;
  double[] real_arrays_originals; // float, double
  int real_arrays_compacted_index;

  private final void cull_compact_resize_real_arrays_log() {
    /* remove duplicates */
    for(int i = 0; i <= real_arrays_index; i++) {
      if(real_arrays_arrays[i] == null) continue;
      for(int n=Math.max(real_arrays_compacted_index,i+1); 
          n<=real_arrays_index; n++) {
        if(real_arrays_arrays[n] == null) continue;
        if((real_arrays_arrays[i] == real_arrays_arrays[n]) && 
           (real_arrays_subscript[i] == real_arrays_subscript[n])) {
          real_arrays_arrays[n] = null;
          real_arrays_subscript[n] = -1;
        }
      }
    }
    /* compact */
    compact: {
      int pointer = real_arrays_index;
      real_arrays_index = real_arrays_compacted_index;
      for(; real_arrays_index <= pointer; real_arrays_index++) {
        if(real_arrays_arrays[real_arrays_index] != null) 
          continue;
        
        while(real_arrays_arrays[pointer] == null && 
              pointer > real_arrays_index) 
          pointer--;
        
        if(real_arrays_index >= pointer)
          break compact;
        
        real_arrays_arrays[real_arrays_index] = 
          real_arrays_arrays[pointer];
        real_arrays_subscript[real_arrays_index] = 
          real_arrays_subscript[pointer];
        real_arrays_originals[real_arrays_index] = 
          real_arrays_originals[pointer];
        real_arrays_arrays[pointer] = null;
        real_arrays_subscript[pointer] = -1;
      }
      break compact;
    }
    real_arrays_compacted_index = real_arrays_index--;
    /* resize */
    if((real_arrays_index+1) <= (real_arrays_arrays.length/2))
      return;

    real_arrays_arrays = resize_object_arr(real_arrays_arrays);
    real_arrays_subscript = resize_int_arr(real_arrays_subscript);
    double[] originals = new double[real_arrays_originals.length << 1];
    System.arraycopy(real_arrays_originals, 0, originals, 0, real_arrays_originals.length);
    real_arrays_originals = originals;   
  }

  private final void rollback_real_arrays_log() {
    int i = real_arrays_index;
    for(;i >= 0; i--) {
      Object arr  = real_arrays_arrays[i];
      int    sub  = real_arrays_subscript[i];
      double orig = real_arrays_originals[i];
      if(arr instanceof float[]) ((float[])arr)[sub]  = (float)orig;
      else                       ((double[])arr)[sub] = orig;
    }
    real_arrays_index = -1;
    real_arrays_compacted_index = 0;
  }

  public final void log(Object array, int subscript, double orig) {
    if(real_arrays_index == real_arrays_arrays.length - 1)
      cull_compact_resize_real_arrays_log();

    real_arrays_arrays[++real_arrays_index] = array;
    real_arrays_subscript[real_arrays_index] = subscript;
    real_arrays_originals[real_arrays_index] = orig;
  }

  /******** Logging Something *********/  
  int something_index;
  UndoSomething[] something_undoers;
  public final void log(UndoSomething u) {
    if(something_index == something_undoers.length - 1) {
      UndoSomething[] undoers = new UndoSomething[something_undoers.length << 1];
      System.arraycopy(something_undoers, 0, undoers, 0, something_undoers.length);
      something_undoers = undoers;
    }
    something_undoers[++something_index] = u;
  }

  private final void rollback_somethings_log() {
    int i = something_index;
    for(;i>=0; i--)
      something_undoers[i].undo();
  }

  private static final int LOG_SIZE = AConstants.__LOG_SIZE__;

  private void set_up_logs() {
    object_index = -1;
    object_compacted_index = 0;
    object_containers = new Object[LOG_SIZE];
    object_undoers = new UndoObject[LOG_SIZE];
    object_originals = new Object[LOG_SIZE];  

    integer_index = -1;
    integer_compacted_index = 0;
    integer_containers = new Object[LOG_SIZE];
    integer_undoers = new UndoInteger[LOG_SIZE];
    integer_originals = new long[LOG_SIZE];  

    real_index = -1;
    real_compacted_index = 0;
    real_containers = new Object[LOG_SIZE];
    real_undoers = new UndoReal[LOG_SIZE];
    real_originals = new double[LOG_SIZE];  

    object_arrays_index = -1;
    object_arrays_compacted_index = 0;
    object_arrays_arrays = new Object[LOG_SIZE][];
    object_arrays_subscript = new int[LOG_SIZE];
    object_arrays_originals = new Object[LOG_SIZE];

    integer_arrays_index = -1;
    integer_arrays_compacted_index = 0;
    integer_arrays_arrays = new Object[LOG_SIZE];
    integer_arrays_subscript = new int[LOG_SIZE];
    integer_arrays_originals = new long[LOG_SIZE];

    real_arrays_index = -1;
    real_arrays_compacted_index = 0;
    real_arrays_arrays = new Object[LOG_SIZE];
    real_arrays_subscript = new int[LOG_SIZE];
    real_arrays_originals = new double[LOG_SIZE];

    something_index = -1;
    something_undoers = new UndoSomething[LOG_SIZE];
  }

  final public void rollback() {
    rollback_objects_log();
    rollback_integers_log();
    rollback_reals_log();
    rollback_object_arrays_log();
    rollback_integer_arrays_log();
    rollback_real_arrays_log();
    rollback_somethings_log();
  }

  public boolean rollback = false; // are we currently in rollback-throwing

  public AThread() { super(); }
  public AThread(Runnable target) { super(target); }
  public AThread(Runnable target, String name) { super(target, name); }
  public AThread(String name) { super(name); }
  public AThread(ThreadGroup group, Runnable target) { super(group, target); }
  public AThread(ThreadGroup group, Runnable target, String name) { super(group, target, name);}
  public AThread(ThreadGroup group, Runnable target, String name, long stackSize) 
  { super(group, target, name, stackSize); }
  public AThread(ThreadGroup group, String name) { super(group, name); } 

  public AThread(AConstructorFormal acons) { super(); }
  public AThread(AConstructorFormal acons, Runnable target) { super(target); }
  public AThread(AConstructorFormal acons, Runnable target, String name) 
  { super(target, name); }
  public AThread(AConstructorFormal acons, String name) { super(name); }
  public AThread(AConstructorFormal acons, ThreadGroup group, Runnable target) 
  { super(group, target); }
  public AThread(AConstructorFormal acons, ThreadGroup group, 
                 Runnable target, String name) 
  { super(group, target, name);}
  public AThread(AConstructorFormal acons, ThreadGroup group, 
                 Runnable target, String name, long stackSize) 
  { super(group, target, name, stackSize); }
  public AThread(AConstructorFormal acons, ThreadGroup group, String name) 
  { super(group, name); }

  /**********************************************************************/ 
  /*   overriding currentThread (called at a furious rate in programs)  */
  /**********************************************************************/ 

  private Thread myThread = null; // almost always null (see below)
//   private static Map<Thread,AThread> evilThreads = 
//     Collections.synchronizedMap(new HashMap<Thread,AThread>());
  private static Map evilThreads = 
    Collections.synchronizedMap(new HashMap());

  static public AThread currentThread() {
    Thread t = Thread.currentThread();
    //    if (t instanceof AThread)
    try {
      return (AThread) t; 
    } catch (ClassCastException e) { 
      AThread ans = (AThread)evilThreads.get(t);
      if(ans != null)
        return ans;
      // would be nice to inform the AtomJava developers when this happens!
      ans = new AThread(); // not for being run!
      ans.set_up_logs(); // since we will never call start deliberately
      evilThreads.put(t,ans);
      ans.myThread = t;
      return ans;
    }
  }
  

  private boolean isThreadAlive() {
    if(myThread != null) return myThread.isAlive();
    return isAlive();
  }

  /**********************************************************************/ 
  /*                       polling for release                          */
  /**********************************************************************/ 
    
  public final void check_for_release() {
    synchronized(this.lock) {
      if(!doomed && lks_to_release.isEmpty())
        return;
    }
    if(in_atomic)
      throw AtomJavaRollBack.ROLLBACK;
    release_lks();
  }

  private static final int CHECK_FREQUENCY = AConstants.__CHECK_FREQUENCY__;

  private int release_count = CHECK_FREQUENCY;
  public final void check_release_maybe() {
    if(--release_count == 0) {
      release_count = CHECK_FREQUENCY;
      check_for_release();
    }
  }

  /**********************************************************************/ 
  /*  lock acquire and release                                          */
  /**********************************************************************/ 

  public final Object lock = new Object(); // guards next five fields
  private boolean in_atomic = false;
  public  boolean doomed    = false;
  public  boolean revocable = false;
  public boolean robbed    = false;

  // TODO: implement a LockSet which includes splaying (such that repeated
  // requests don't suffer O(n) more then once.
  private class LockSet {
    AObject[] locks = new AObject[AConstants.__LOCKSET_SIZE__];
    int index = 0;

    void addLock(AObject lk) {
      for(int i = 0; i < index; i++)
        if(locks[i] != lk) continue;
        else return;
      
      if(index == locks.length) {
        AObject[] tmp = new AObject[locks.length << 1];
        System.arraycopy(locks, 0, tmp, 0, locks.length);
        locks = tmp;
      }

      locks[index++] = lk;
    }

    void releaseLocks() {
      for(int i = 0; i < index; i++) {
        rel_lk(locks[i]);
        locks[i] = null;
      }
      index = 0;
    }

    boolean isEmpty() { return index == 0; }
  }

  // guarded by lock above ; LockSet is otherwise NOT thread safe
  private final LockSet lks_to_release = new LockSet();

  //  private HashSet<AObject> lks_to_release = new HashSet<AObject>();
    
  // these locks guard the AtomJava fields in AObjects  
  private final static int LOCKS_SIZE = 1 << 10;
  private static final Object[] locks = new Object[LOCKS_SIZE];
  static { 
    for(int i = 0; i < locks.length; i++)
      locks[i] = new Object();
  }

  public final void rel_lk(AObject lk) {
    Object obj_lock = lk; //locks[System.identityHashCode(lk) % LOCKS_SIZE];
    synchronized(obj_lock) {  // CAS(lk.currentHolder, this, null);
      if(lk.currentHolder == this)
        lk.currentHolder = null;
    }
  }
    
  private final void release_lks() {
    synchronized(this.lock) {
      lks_to_release.releaseLocks();
//       for(Iterator<AObject> i = lks_to_release.iterator(); i.hasNext();)
//         rel_lk(i.next());
//      lks_to_release.clear();
    }
  }

  // helper method for acq_lk and acq_arraylk
  private static final boolean try_to_steal(AThread revokee, AObject lk) {
    synchronized(revokee.lock) { // serves as a barrier between threads
      if(revokee.revocable) { 
        if(revokee.in_atomic) {
          revokee.robbed = true;
          revokee.rollback();
        }
        revokee.rel_lk(lk);
        return true;
      }
        
      if(!revokee.isThreadAlive()) { // TO DO: what if in atomic? rollback?
        revokee.rel_lk(lk);
        return true;
      }

      revokee.lks_to_release.addLock(lk);
      return false;
    }
  }

  public final void acq_lk(AObject lk) { 
    Object obj_lock = lk; //locks[System.identityHashCode(lk) % LOCKS_SIZE];
    AThread revokee;
    while(true) {
      synchronized(obj_lock) {
        if(lk.currentHolder == null) {// CAS(lk.currentHolder, null, this)
          lk.currentHolder = this;
          return;
        }
        revokee = lk.currentHolder;
      }

      check_for_release();
            
      // TO DO: class initialization deadlock detection/resolution
      // TO DO: should we allow lock stealing (see aj_yield)
      if(!try_to_steal(revokee,lk)) // do not yield if we revoked
        Thread.yield();
    }
  }

  // these locks guard elements of 'array' below
  private static final int ARRAY_LOCKS_SIZE = 1 << 10;
  private static AObject[] array_locks = new AObject[ARRAY_LOCKS_SIZE];
  static {
    for(int i = 0; i < array_locks.length; i++)
      array_locks[i] = new AObject((AThread)null);
  }
    
  // TO DO: we don't synchronize on array_locks AT ALL, this might not be 
  //        kosher come dynamic locking
  final static public void acq_arraylk(Object array) {
    acq_arraylk(AThread.currentThread(),array);
  }
  final static public void acq_arraylk(AThread me, Object array) {
    int hash = System.identityHashCode(array) % ARRAY_LOCKS_SIZE;
    AObject lk = array_locks[hash];
    if(lk.currentHolder == me)
      return;
    Object obj_lock = lk; // locks[System.identityHashCode(lk) % LOCKS_SIZE];    
    while(true) {
      AThread revokee;
      synchronized(obj_lock) {
        if(lk.currentHolder == null) {// CAS(lk.currentHolder, null, this)
          lk.currentHolder = me;
          break;
        }
        revokee = lk.currentHolder;
      }

      me.check_for_release();
            
      // TO DO: class initialization deadlock detection/resolution
      // TO DO: should we allow lock stealing (see aj_yield)
      if(!try_to_steal(revokee,lk)) // do not yield if we revoked
        Thread.yield();
    }

    if(lk != array_locks[hash]) { // needed only when locks change dynamically
      me.rel_lk(lk);
      me.check_for_release();
      acq_arraylk(me,array);
    }
  }
    
  /**********************************************************************/ 
  /*  entering and leaving atomic                                       */
  /**********************************************************************/ 
    
  public final void start_atomic() { 
    in_atomic = true;
  }

  public final void end_atomic() { 
    in_atomic = doomed = false;
    rollbacks = 0;

    // TO DO: potentially shorten arrays

    object_index = -1;
    integer_index = -1;
    real_index = -1;
    object_arrays_index = -1;
    integer_arrays_index = -1;
    real_arrays_index = -1;
    something_index = -1;

    object_compacted_index = 0;
    integer_compacted_index = 0;
    real_compacted_index = 0;
    object_arrays_compacted_index = 0;
    integer_arrays_compacted_index = 0;
    real_arrays_compacted_index = 0;

  }

  private int rollbacks = 0;
  private static final int SLEEP_TIME = 10;

  final public void rollback_and_sleep() {
    rollback();
    release_lks(); // why not?
    try { sleep(SLEEP_TIME * ++rollbacks); } // TODO: sleep set revocable?
    catch (InterruptedException ie) {}
  }

  /*********************************************************************/
  /*  Acting like an AObject (we extend Thread, not AObject)           */
  /*********************************************************************/

  public AObject currentLock = new AObject(this);

  public final void acq_mylk() {
    acq_mylk(currentThread());
  }

  public final void acq_mylk(AThread me) { // acq my current lock (copied from AObject)
    //System.out.println("AThread.acq_mylk()");
    AObject lk = currentLock;
    if(currentLock.currentHolder == me)
      return;
    //System.out.println("currentLock.currentHolder != me");
    me.acq_lk(lk);
    if(lk != currentLock) { // needed only when locks change dynamically
      //System.out.println("lk != currentLock");
      me.rel_lk(lk);  // unnecessary but a good idea?
      me.check_for_release(); // just to be careful because in a loop
      acq_mylk(me); // very unlikely tail-call
    }
  }

  /**********************************************************************/ 
  /*                  handling class intialization                      */
  /**********************************************************************/ 

  private long history = 0;
  static final long HISTORY_OVERFLOW = 1 << 63;

  final public void enter_clinit() {
    if((history & HISTORY_OVERFLOW) == HISTORY_OVERFLOW)
      throw new Error("nested atomic class init overflow");
    rollback();
    history = (history << 1) | (in_atomic ? 1 : 0);
    in_atomic = false;
  }

  final public void exit_clinit() {
    doomed = in_atomic = ((1 & history) == 1);
    history >>>= 1;
  }

  /***********************************************************************/
  /*          _____ _ _  _  ___ ___ __      ___ ___ _    _ ___           */
  /*            |   |_| |_) |_  |_| | |      |  |_|  \  /  |_|           */
  /*            |   | | | \ |__ | | |_| o (__|  | |   \/   | |           */
  /* override and provide atomic versions of methods in Thread           */
  /***********************************************************************/
  
  public final String __aj_mangle_getName() {
    acq_mylk();
    return getName();
  }
  public final String __aj_atomic___aj_mangle_getName() {
    acq_mylk();
    return getName();
  }
  public final void __aj_mangle_setName(String name) {
    acq_mylk();
    setName(name);
  }
  public final void __aj_atomic___aj_mangle_setName(String name) {
    acq_mylk();
    currentThread().log(this, __aj_undo_name_, getName());
    setName(name);
  }
  private static UndoObject __aj_undo_name_ = 
    new UndoObject() {
      public void undo(Object o, Object v) { ((Thread)o).setName((String)v); };
    };
    
  public final int __aj_mangle_getPriority() {
    acq_mylk();
    return getPriority();
  }
  public final int __aj_atomic___aj_mangle_getPriority() {
    acq_mylk();
    return getPriority();
  }
  public final void __aj_mangle_setPriority(int newPriority) {
    acq_mylk();
    setPriority(newPriority);
  }
  public final void __aj_atomic___aj_mangle_setPriority(int newPriority) {
    acq_mylk();
    currentThread().log(this, __aj_undo_priority_, getPriority());
    setPriority(newPriority);
  }
  private static UndoInteger __aj_undo_priority_ = 
    new UndoInteger() {
      public void undo(Object o, long v) { ((Thread)o).setPriority((int)v); };
    };
    

  public final boolean __aj_mangle_isDaemon() {
    acq_mylk();
    return isDaemon();
  }
  public final boolean __aj_atomic___aj_mangle_isDaemon() {
    acq_mylk();
    return isDaemon();
  }
  public final void __aj_mangle_setDaemon(boolean on) {
    acq_mylk();
    setDaemon(on);
  }
  public final void __aj_atomic___aj_mangle_setDaemon(boolean on) {
    acq_mylk();
    currentThread().log(this, __aj_undo_daemon_, isDaemon() ? 1 : 0);
    setDaemon(on);
  }
  private static UndoInteger __aj_undo_daemon_ = 
    new UndoInteger() {
      public void undo(Object o, long v) {  ((Thread)o).setDaemon(v == 1 ? true : false); };
    };
    

  public final ThreadGroup __aj_mangle_getThreadGroup() {
    acq_mylk();
    return getThreadGroup();
  }
  public final ThreadGroup __aj_atomic___aj_mangle_getThreadGroup() {
      acq_mylk();
      return getThreadGroup();
  } 
  public ClassLoader __aj_mangle_getContextClassLoader() {
    acq_mylk();
    return getContextClassLoader();
  }
  public ClassLoader __aj_atomic___aj_mangle_getContextClassLoader() {
    acq_mylk();
    return getContextClassLoader();
  }
  public void __aj_mangle_setContextClassLoader(ClassLoader cl) {
    acq_mylk();
    setContextClassLoader(cl);
  }
  public void __aj_atomic___aj_mangle_setContextClassLoader(ClassLoader cl) {
    acq_mylk();
    currentThread().log(this, __aj_undo_contextClassLoader_, getContextClassLoader());
    setContextClassLoader(cl);
  }
  private static UndoObject __aj_undo_contextClassLoader_ = 
    new UndoObject() {
      public void undo(Object o, Object v) {  ((Thread)o).setContextClassLoader((ClassLoader)v); };
    };

  public static void yield() {
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { Thread.yield(); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public static void __aj_atomic_yield() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  
  public static void sleep(long millis) throws InterruptedException {
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { Thread.sleep(millis); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public static void __aj_atomic_sleep(long millis) throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }
  
    
  public static void sleep(long millis, int nanos) throws InterruptedException {
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { Thread.sleep(millis, nanos); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public static void __aj_atomic_sleep(long millis, int nanos) throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public synchronized void start() {
    set_up_logs(); super.start();
  }
  public synchronized void __aj_atomic_start() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public void __aj_atomic_run() {
     super.run();
  }
    
  public final void __aj_mangle_stop()                 {throw new NoSuchMethodError();}
  public final void __aj_mangle_stop(Throwable o)      {throw new NoSuchMethodError();}
  public final void __aj_atomic___aj_mangle_stop()           {throw new NoSuchMethodError();}
  public final void __aj_atomic___aj_mangle_stop(Throwable o){throw new NoSuchMethodError();}

  public void __aj_atomic_interrupt() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public static boolean __aj_atomic_interrupted() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public boolean __aj_atomic_isInterrupted() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public void __aj_atomic_destroy() { throw new NoSuchMethodError(); }
  // TO DO? Why 1 of these, but 2 of stop versions?
  // Answer: Because the non-atomic version (from Thread) does the right thing -ben

  public final boolean __aj_atomic_isAlive() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }

  public final void __aj_mangle_suspend() { throw new NoSuchMethodError();}
  public final void __aj_atomic___aj_mangle_suspend() { 
    throw new NoSuchMethodError();
  }
  public final void __aj_mangle_resume() { throw new NoSuchMethodError(); }
  public final void __aj_atomic___aj_mangle_resume() { 
    throw new NoSuchMethodError(); 
  }

  public static int activeCount() {
    return currentThread().__aj_mangle_getThreadGroup().activeCount();
  }
  public static int __aj_atomic_activeCount() {
    return activeCount();
  }

  public static int enumerate(Thread tarray[]) {
    return currentThread().__aj_mangle_getThreadGroup().enumerate(tarray);
  }
  public static int __aj_atomic_enumerate(Thread tarray[]) {
    return enumerate(tarray);
  }

  public int __aj_atomic_countStackFrames() {
    throw new AtomJavaNativeInAtomicError("can\'t call native methods in atomic!");
  }
    
  // since we might not run for a long time, let others steal our locks.
  // Note: these may NOT be called from in an atomic (which we can
  // 'enforce' by not making atomic versions)
  // grr Thread.join is final!!
  public void __aj_mangle_join() throws InterruptedException { 
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.join(); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public void __aj_mangle_join(long m) throws InterruptedException {
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.join(m); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public void __aj_mangle_join(long m, int n) throws InterruptedException {
    AThread t = currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { super.join(m,n); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
    
  public void __aj_atomic___aj_mangle_join() throws InterruptedException { 
    throw new AtomJavaSleepInAtomicException("join");
  }
  public void __aj_atomic___aj_mangle_join(long m) throws InterruptedException { 
    throw new AtomJavaSleepInAtomicException("join");
  }
  public void __aj_atomic___aj_mangle_join(long m, int n) throws InterruptedException { 
    throw new AtomJavaSleepInAtomicException("join");
  }

  public static void __aj_atomic_dumpStack() { // TO DO? just call dumpStack()?
    new Exception("Stack trace").printStackTrace();
  }

  public final void __aj_atomic_checkAccess() {
    super.checkAccess();
  }    
    
  // could just call toString() but until group.getName is assured to have no side effects, hold
  public String __aj_atomic_toString() {
    if ( __aj_atomic___aj_mangle_getThreadGroup() != null)
      return "Thread[" + __aj_atomic___aj_mangle_getName() + 
        "," + __aj_atomic___aj_mangle_getPriority() + "," + 
        __aj_atomic___aj_mangle_getThreadGroup().getName() + "]";
    return "Thread[" + __aj_atomic___aj_mangle_getName() + 
      "," + __aj_atomic___aj_mangle_getPriority() + ",]";
  }

  public int     __aj_atomic_hashCode()         { return super.hashCode();  }
  public boolean __aj_atomic_equals(Object obj) { return super.equals(obj); }
  public Class   __aj_atomic_getClass()         { return super.getClass();  }
  protected void __aj_atomic_finalize() throws Throwable { super.finalize(); }

  public void __aj_atomic_notify() {
    throw new AtomJavaNativeInAtomicError("notify");
  }
  public void __aj_atomic_notifyAll() {
    throw new AtomJavaNativeInAtomicError("notifyAll");
  }

  protected Object clone() throws CloneNotSupportedException {
    acq_mylk();
    return super.clone();
  }
  protected Object __aj_atomic_clone() throws CloneNotSupportedException {
    acq_mylk();
    return super.clone();
  }

  public final void __aj_atomic___aj_wait()
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
  public final void __aj_atomic___aj_wait(long timeout, int nanos) 
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
  public final void __aj_atomic___aj_wait(long timeout) 
    throws InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }
}

class AtomJavaSleepInAtomicException extends Error {
  public AtomJavaSleepInAtomicException(String message) {
    super(message);
  }
}
class NewWithFinalizerInAtomic extends Error {}
