package polyglot.ext.atomjava.runtime.wrapper.java.lang;

import polyglot.ext.atomjava.runtime.*;

  /**************************************************************************/
  /*       static atomic wrappers for static type java.lang.Object          */
  /**************************************************************************/

// TO DO: deal with Throwable

/* NOTES:

  It is most definately possible to reenter an atomic scope if we call the 
  non-atomic version (in place of the atomic version), however, we are hoping
  that will not happen.

*/

public class Object {

  // everything now gets a wrapper, so even though clone and finalize are
  // protected, something, like Throwable, might not be a subclass of 
  // an AObject and therefore need this wrapper! (Ben wrote this comment
  // and now Ben is reading it and Ben still doesn't see how clone or finalize
  // can get called if we have static type Object because these methods
  // are protected!)

  public static final java.lang.Object clone(java.lang.Object o) 
    throws java.lang.CloneNotSupportedException {
    throw new java.lang.NoSuchMethodError(o + ".clone");
  }

  public static final java.lang.Object __aj_atomic_clone(java.lang.Object o) 
    throws java.lang.CloneNotSupportedException {
    throw new java.lang.NoSuchMethodError(o + ".__aj_atomic_clone");
  }

  public static final void finalize(java.lang.Object o) 
    throws java.lang.Throwable {
    throw new java.lang.NoSuchMethodError(o + ".finalize");
  }
  public static final void __aj_atomic_finalize(java.lang.Object o) 
    throws java.lang.Throwable {
    throw new java.lang.NoSuchMethodError(o + ".__aj_atmomic_finalize");
  }

  static final public boolean equals(java.lang.Object o, java.lang.Object obj) {
    return o.equals(obj); // DAN: always safe I hope! 
  }
  static final public boolean __aj_atomic_equals(java.lang.Object o, 
                                                 java.lang.Object obj) {
    if(o instanceof AObject)
      return ((AObject)o).__aj_atomic_equals(obj);
    if(o instanceof AThread)
      return ((AThread)o).__aj_atomic_equals(obj);

    if(o.getClass().isArray())
      return o.equals(obj); // TODO: IF THIS DOES STRUCTURAL EQUALITY SHALL WE LOCK IT?

    /* dynamic type uknown */
    throw new java.lang.NoSuchMethodError(o + ".__aj_atomic_equals_");
  }

  static final public Class getClass(java.lang.Object o) {
    return o.getClass(); /* getClass is final ;) */
  }  
  static final public Class __aj_atomic_getClass(java.lang.Object o) {
    return o.getClass(); /* getClass is final ;) */
  }

  static final public int hashCode(java.lang.Object o) {
    return o.hashCode(); /* DAN: always safe I hope! */
  }
  static final public int __aj_atomic_hashCode(java.lang.Object o) {
    if(o instanceof AObject)
      return ((AObject)o).__aj_atomic_hashCode();
    if(o instanceof AThread) 
      return ((AThread)o).__aj_atomic_hashCode();

    if(o.getClass().isArray())
      return o.hashCode();
    
    /* dynamic type unknown */
    throw new java.lang.NoSuchMethodError(o + ".__aj_atomic_hashCode_");
  }

  static final public void notify(java.lang.Object o) {
    o.notify();
  }
  static final public void __aj_atomic_notify(java.lang.Object o) {
    throw new AtomJavaNativeInAtomicError("notify");
  }

  static final public void notifyAll(java.lang.Object o) {
    o.notify();
  }
  static final public void __aj_atomic_notifyAll(java.lang.Object o) {
    throw new AtomJavaNativeInAtomicError("notifyAll");    
  }
  
  static final public java.lang.String toString(java.lang.Object o) {
    return o.toString(); /* DAN: always safe I hope */
  }
  static final public java.lang.String __aj_atomic_toString(java.lang.Object o) {
    if(o instanceof AObject)
      return ((AObject)o).__aj_atomic_toString();
    if(o instanceof AThread)
      return ((AThread)o).__aj_atomic_toString();
      
    if(o.getClass().isArray())
      return o.toString();
    
    /* dynamic type uknown */
    throw new java.lang.NoSuchMethodError(o + ".__aj_atomic_toString_");
  }

  
  public static final void wait(java.lang.Object o)
    throws java.lang.InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { o.wait(); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }  
  public static final void __aj_atomic_wait(java.lang.Object o)
    throws java.lang.InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }

  public static final void wait(java.lang.Object o, 
                                     long timeout, 
                                     int nanos) 
    throws java.lang.InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { o.wait(timeout, nanos); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public static final void __aj_atomic_wait(java.lang.Object o, 
                                                        long timeout, 
                                                        int nanos) 
    throws java.lang.InterruptedException {
     throw new AtomJavaNativeInAtomicError("wait");
  }

  public static final void wait(java.lang.Object o, 
                                     long timeout) 
    throws java.lang.InterruptedException {
    AThread t = AThread.currentThread();
    synchronized(t.lock) { t.revocable = true; }
    try { o.wait(timeout); }
    finally { synchronized(t.lock) { t.revocable = false; } }
  }
  public static final void __aj_atomic_wait(java.lang.Object o, 
                                                 long timeout) 
    throws java.lang.InterruptedException {
    throw new AtomJavaNativeInAtomicError("wait");
  }

}
