package polyglot.ext.atomjava.types;

import polyglot.types.*;
import polyglot.main.*;
import polyglot.frontend.*;
import polyglot.ext.jl.types.TypeSystem_c;
import polyglot.ext.atomjava.util.*;
import polyglot.util.*;
import java.util.*;

public class AtomJavaTypeSystem_c extends TypeSystem_c implements AtomJavaTypeSystem {

  protected ExtensionInfo ext;

  public AtomJavaTypeSystem_c(ExtensionInfo ext) { this.ext = ext; }

  public Context createContext() {
    return new AtomJavaContext_c(this);
  }

  protected ClassType THREAD_;
  protected ClassType ATHREAD_;
  protected ClassType AOBJECT_;
  protected ClassType AARRAY_;
  protected ClassType AMAIN_;
  protected ClassType ACONS_;
  protected ClassType NEWWITHFINALIZER_;
  protected ClassType AJNATIVEINATOMICE_;
  protected ClassType AJUNREACHABLE_;
  protected ClassType AJROLLBACK_;
  protected ClassType UNDOINTEGER_;
  protected ClassType UNDOREAL_;
  protected ClassType UNDOOBJECT_;
  protected ClassType runtime_java_lang_Object_;
  protected ClassType FAILURE_;
  protected ClassType NOSUCHMETHODERROR_;

  private final ClassType runtime_load(String s) {
    return load(RUNTIME + s);
  }

  public ClassType NoSuchMethodError() {
    if(NOSUCHMETHODERROR_ != null) return NOSUCHMETHODERROR_;
    return NOSUCHMETHODERROR_ = load("java.lang.NoSuchMethodError");
  }

  public ClassType Failure() {
    if(FAILURE_ != null) return FAILURE_;
    return FAILURE_ = runtime_load("Failure");
  }

  public ClassType Thread() {
    if(THREAD_ != null) return THREAD_;
    return THREAD_ = load("java.lang.Thread");
  }

  public ClassType runtime_java_lang_Object() {
    if(runtime_java_lang_Object_ != null) return runtime_java_lang_Object_;
    return runtime_java_lang_Object_ = runtime_load("wrapper.java.lang.Object");
  }

  public ClassType AThread() { 
    if (ATHREAD_ != null) return ATHREAD_;
    return ATHREAD_ = runtime_load("AThread"); 
  }
  public ClassType AObject() { 
    if (AOBJECT_ != null) return AOBJECT_;
    return AOBJECT_ = runtime_load("AObject"); 
  }
  public ClassType AArray() { 
    if (AARRAY_ != null) return AARRAY_;
    return AARRAY_ = runtime_load("AArray"); 
  }
  public ClassType AMain() { 
    if (AMAIN_ != null) return AMAIN_;
    return AMAIN_ = runtime_load("AMain"); 
  }
  public ClassType AtomJavaNativeInAtomicError() {
    if (AJNATIVEINATOMICE_ != null) return AJNATIVEINATOMICE_;
    return AJNATIVEINATOMICE_ = runtime_load("AtomJavaNativeInAtomicError");
  }
  public ClassType AtomJavaUnreachableCodeError() {
    if (AJUNREACHABLE_ != null) return AJUNREACHABLE_;
    return AJUNREACHABLE_ = runtime_load("AtomJavaUnreachableCodeError");
  }
  public ClassType AtomJavaRollBack() {
    if (AJROLLBACK_ != null) return AJROLLBACK_;
    return AJROLLBACK_ = runtime_load("AtomJavaRollBack");
  }
  public ClassType AConstructorFormal() { 
    if (ACONS_ != null) return ACONS_;
    return ACONS_ = runtime_load("AConstructorFormal"); 
  }
  public ClassType NewWithFinalizerInAtomic() { 
    if (NEWWITHFINALIZER_ != null) return NEWWITHFINALIZER_;
    return NEWWITHFINALIZER_ = runtime_load("NewWithFinalizerInAtomic"); 
  }

  public ClassType undoerType(Type t) {
    if(t.isBoolean() || t.isByte() || t.isChar() || 
       t.isShort() || t.isInt() || t.isLong())
      if(UNDOINTEGER_ != null) return UNDOINTEGER_;
      else return UNDOINTEGER_ = runtime_load("UndoInteger");
    if(t.isFloat() || t.isDouble())
      if(UNDOREAL_ != null) return UNDOREAL_;
      else return UNDOREAL_ = runtime_load("UndoReal");
    if(UNDOOBJECT_ != null) return UNDOOBJECT_;
    return UNDOOBJECT_ = runtime_load("UndoObject");
  }

  public Type undoValueType(Type t) {
    if(t.isBoolean() || t.isByte() || t.isChar() || 
       t.isShort() || t.isInt() || t.isLong())
      return Long();
    if(t.isFloat() || t.isDouble())
      return Double();
    return Object();
  }

  private HashMap<String,ClassType> type_updates = 
    new HashMap<String,ClassType>();

  public void addTypeUpdate(ClassType from, ClassType to) {
    if(from.fullName() == null)
      throw new InternalCompilerError("ClassType fullName required " +
                                      "for making type updates", 
                                      from.position());

    /* update the system resolver for any subsequent find's/typeForNames's */
    /* note: we only need to update the systemResolver because all requests
       for types are routed through this TopLevelResolver first */
    try {
      ((CachingResolver) systemResolver()).addNamed(to.fullName(), to);
    } catch (SemanticException e) { 
      throw new InternalCompilerError(e);
    }
    
    type_updates.put(from.fullName(), to);
  }

  public void addTypeUpdate(ClassType t) {
    addTypeUpdate(t, t);
  }

  public Type updateType(Type t) {
    if(t == null) return t;
    
    if(t.isArray()) return t.toArray().base(updateType(t.toArray().base()));

    if(!t.isClass()) return t;

    if(equals(t, Thread())) return AThread();

    Type nt = (Type) type_updates.get(t.toClass().fullName());

    return (nt == null) ? t : nt;
  }

  public List updateTypes(List list) {
    List l = new LinkedList();
    for(Iterator i = list.iterator(); i.hasNext();)
      l.add(updateType((Type)i.next()));
    return l;
  }

  /* update*Instance methods assumes that you have included any new pieces 
     of information (formal's, throw's, etc) however that you might not have
     accurate type information 
  */

  public MethodInstance updateMethodInstance(MethodInstance mi) {
    mi = mi.container((ReferenceType)updateType(mi.container()));
    mi = mi.returnType(updateType(mi.returnType()));
    mi = mi.formalTypes(updateTypes(mi.formalTypes()));
    mi = mi.throwTypes(updateTypes(mi.throwTypes()));
    return mi;
  }

  public ConstructorInstance updateConstructorInstance(ConstructorInstance ci) {
    ci = ci.container((ClassType)updateType(ci.container()));
    ci = ci.formalTypes(updateTypes(ci.formalTypes()));
    ci = ci.throwTypes(updateTypes(ci.throwTypes()));
    return ci;
  }

  public LocalInstance updateLocalInstance(LocalInstance li) {
    li = li.type(updateType(li.type()));
    return li;
  }

  public FieldInstance updateFieldInstance(FieldInstance fi) {
    fi = fi.container((ReferenceType)updateType(fi.container()));
    fi = fi.type(updateType(fi.type()));
    return fi;
  }

  public InitializerInstance updateInitializerInstance(InitializerInstance ii) {
    ii = ii.container((ClassType)updateType(ii.container()));
    return ii;
  }

  public boolean useWrappers(ClassType t) {
    return !isSubtype(t,AObject()) && !isSubtype(t,AThread());
  }
}
