package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.main.AJOptions;
import polyglot.ext.atomjava.util.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

public class AJSignatureBuilder extends AJContextVisitor {
  
  public AJSignatureBuilder(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  // accumulated field decls for innerclasses
  private static List field_decls = new LinkedList();

  public Node leave(Node parent, Node old, Node n, NodeVisitor v) {

    if(!(n instanceof ClassBody))
       return n;

    ClassBody cb = (ClassBody)n;

    /* clear out STATIC_LOCK_HOLDER */
    STATIC_LOCK_HOLDER = null;

    /* members to add to the current list of members */
    List members = new LinkedList();
    
    /* the class type of the class body */
    ParsedClassType ct = ((AJContextVisitor)v).context().currentClassScope();

    for (Iterator i = cb.members().iterator(); i.hasNext(); ) {
      ClassMember member = (ClassMember) i.next();
      
      if(member instanceof MethodDecl)
        build_MethodDecl((MethodDecl)member, members, 
                         ct, (AJContextVisitor)v);
      
      if(member instanceof ConstructorDecl)
        build_ConstructorDecl((ConstructorDecl)member, members, 
                              ct, (AJContextVisitor)v);

      if(member instanceof FieldDecl)
        build_FieldDecl((FieldDecl)member, members, 
                        ct, (AJContextVisitor)v);
    }

    /* HACK: create setters/getters for field_decls in inner class */
    if(!ct.isInnerClass() && !field_decls.isEmpty()) {
      for(Iterator i = field_decls.iterator(); i.hasNext();)
	build_InnerClassFieldDecl((FieldDecl)i.next(), members, 
				  ct, (AJContextVisitor)v);
      field_decls = new LinkedList();
    }

    /* declare the type update */
    typeSystem().addTypeUpdate(ct);

    /* combine the list of members */
    members.addAll(cb.members());

    /* update the ClassBody */
    return cb.members(members);
  }

  protected void build_MethodDecl(MethodDecl md, List members, 
                                  ParsedClassType ct, AJContextVisitor v) {
    MethodDecl atomic_md = 
      v.nodeFactory().AtomicMethodDecl(v.context(), md.flags(), md.returnType(), 
                                       v.context().atomicMethodName(md.name()),
                                       md.formals(), md.throwTypes(), md.body());
    /* add the atomic version of the method to the class */
    members.add(atomic_md);
    ct.addMethod(atomic_md.methodInstance());
  }

  protected void build_ConstructorDecl(ConstructorDecl cd, List members,
                                       ParsedClassType ct, AJContextVisitor v) {
    AtomJavaContext     c  = v.context();
    AtomJavaTypeSystem  ts = v.typeSystem();
    AtomJavaNodeFactory nf = v.nodeFactory();
    
    Formal f = nf.Formal(c, Flags.NONE,
                         nf.CanonicalTypeNode(c, ts.AConstructorFormal()), 
                         c.atomicFormalName());

    List formals = new LinkedList(cd.formals());
    formals.add(0,f);

    ConstructorDecl atomic_cd = 
      v.nodeFactory().AtomicConstructorDecl(v.context(), cd.flags(), 
                                            cd.name(), formals, 
                                            cd.throwTypes(), cd.body());

    /* add the atomic version of the constructor as a class change */
    members.add(atomic_cd);
    ct.addConstructor(atomic_cd.constructorInstance());
  }

  public void build_InnerClassFieldDecl(FieldDecl fd, List members,
					ParsedClassType ct, AJContextVisitor v) {
    String name = fd.name();
    ClassType ict = (ClassType) fd.fieldInstance().container();

    for(; ict.isInnerClass(); ict = ict.outer())
	name = ict.name() + "_" + name;

    // assert ict.equals(ct);

    build_FieldDecl(fd.name(name), members, ct, v);
  }

  public void build_FieldDecl(FieldDecl fd, List members,
                              ParsedClassType ct, AJContextVisitor v) {

    if(ct.isInnerClass()) {
      field_decls.add(fd);
      return;
    }

    AtomJavaContext     c  = v.context();
    AtomJavaTypeSystem  ts = v.typeSystem();
    AtomJavaNodeFactory nf = v.nodeFactory();
    
    Flags     fd_flags     = fd.flags();
    Type      fd_type      = fd.declType();
    String    fd_name      = fd.name();
    Type      fd_container = fd.fieldInstance().container();
    
    if(fd_flags.isFinal()) return;
    
    Formal gs_formal_o = 
      nf.Formal(c, Flags.NONE, nf.CanonicalTypeNode(c, fd_container), "o");
    Formal s_formal_v = 
      nf.Formal(c, Flags.NONE, nf.CanonicalTypeNode(c, fd_type), "v");
    Formal u_formal_o = 
      nf.Formal(c, Flags.NONE, nf.CanonicalTypeNode(c, ts.Object()), "o");
    Formal u_formal_v = 
      nf.Formal(c, Flags.NONE, 
                nf.CanonicalTypeNode(c, ts.undoValueType(fd_type)), "v");
    Formal gs_formal_me =
      nf.Formal(c, Flags.NONE, nf.CanonicalTypeNode(c, ts.AThread()), "me");

    Expr undo_left, field_arg, lockobj;
    
    if(fd_flags.isStatic()) {

      if(!members.contains(STATIC_LOCK_HOLDER(c, nf, ts))) {
        members.add(STATIC_LOCK_HOLDER(c, nf, ts));
        ct.addField(STATIC_LOCK_HOLDER(c, nf, ts).fieldInstance());
      }

      undo_left = field_arg = nf.Field(c, fd);
      lockobj = nf.Field(c,STATIC_LOCK_HOLDER(c, nf, ts));
    } else {
      field_arg = nf.Field(c, fd, nf.Local(c, gs_formal_o));
      undo_left = 
        nf.Field(c, fd, nf.Cast(c, nf.CanonicalTypeNode(c, fd_container),
                                nf.Local(c, u_formal_o)));
      lockobj   = nf.Local(c, gs_formal_o);
    }

    Expr acq_lk    = nf.Call(c, lockobj, "acq_mylk", 
                             Collections.EMPTY_LIST);
    Expr acq_lk_me = nf.Call(c, lockobj, "acq_mylk", 
                             AJS.List(nf.Local(c,gs_formal_me)));
    
    Flags get_set_flags = fd_flags.clear(Flags.VOLATILE).set(Flags.STATIC);
    
    /* getter methods: 
       static T get_f(AThread me, C o){ acq_lk(me); return f; } 
       static T get_f(C o){ acq_lk(); return f; } */
    List getter1_formals = AJS.List(gs_formal_me,gs_formal_o);
    List getter2_formals = AJS.List(gs_formal_o);
    List getter1_stmts = AJS.List(nf.Eval(c, acq_lk_me),nf.Return(c,field_arg));
    List getter2_stmts = AJS.List(nf.Eval(c, acq_lk),nf.Return(c,field_arg));
    MethodDecl getter1 = nf.MethodDecl(c, get_set_flags, fd_type, 
                                       c.fieldGetterName(fd_name), 
                                       getter1_formals, getter1_stmts);
    MethodDecl getter2 = nf.MethodDecl(c, get_set_flags, fd_type, 
                                       c.fieldGetterName(fd_name), 
                                       getter2_formals, getter2_stmts);
    members.add(getter1);
    members.add(getter2);
    ct.addMethod(getter1.methodInstance());
    ct.addMethod(getter2.methodInstance());
    
    /* the undoer field and anonymous inner class */
    Expr undo_right = nf.Local(c, u_formal_v);
    if(!fd_type.isBoolean())
      undo_right = nf.Cast(c,nf.CanonicalTypeNode(c,fd_type),undo_right);
    else
      undo_right = nf.Conditional(c, nf.Binary(c, undo_right, Binary.EQ, 
                                               nf.IntLit(c, IntLit.INT, 1),
                                               ts.Int()), 
                                  nf.BooleanLit(c, true), 
                                  nf.BooleanLit(c, false));

    List undo_formals = AJS.List(u_formal_o, u_formal_v);
    List undo_stmts   = 
      AJS.List(nf.Eval(c, nf.Assign(c, undo_left, undo_right)));
    MethodDecl undo_method = nf.MethodDecl(c,Flags.PUBLIC,ts.Void(),"undo", 
                                           undo_formals,undo_stmts);
    
    ClassBody class_body = nf.ClassBody(c, AJS.List(undo_method));
    TypeNode  undo_node  = nf.CanonicalTypeNode(c, ts.undoerType(fd_type));
    New       undo_impl  = nf.New(c, null, undo_node, Collections.EMPTY_LIST, 
                                  class_body);
    FieldDecl undoer = nf.FieldDecl(c, Flags.STATIC.set(Flags.PRIVATE), 
                                    undo_node, c.fieldUndoerName(fd_name), 
                                    undo_impl);

    members.add(undoer);
    ct.addField(undoer.fieldInstance());

    /* atomic setter */
    List setter_formals = AJS.List(gs_formal_o, s_formal_v);
    Expr log_arg = field_arg;
    if(fd_type.isBoolean())
      log_arg = nf.Conditional(c, log_arg, nf.IntLit(c, IntLit.INT, 1), 
                               nf.IntLit(c, IntLit.INT, 0));
    
    List log_arguments  = AJS.List(nf.Local(c, gs_formal_o),
                                   nf.Field(c, undoer),
                                   log_arg);
    List atomic_setter_stmts = 
      AJS.List(nf.Eval(c, acq_lk),
               nf.Eval(c, nf.Call(c, nf.currentThread(c), "log", 
                                  log_arguments)),
               nf.Return(c, nf.Assign(c, field_arg, nf.Local(c, s_formal_v))));
    
    MethodDecl setter_atomic = 
      nf.MethodDecl(c, get_set_flags, fd_type, 
                    ((AtomJavaContext)c.pushAtomic()).fieldSetterName(fd_name),
                    setter_formals, atomic_setter_stmts);
        
    members.add(setter_atomic);
    ct.addMethod(setter_atomic.methodInstance());
    
    /* non-atomic setter */
    if(!((AJOptions)v.job().extensionInfo().getOptions()).weak) {
      List setter_stmts   = 
        AJS.List(nf.Eval(c, acq_lk),
                 nf.Return(c, nf.Assign(c, field_arg,
                                        nf.Local(c, s_formal_v))));
      MethodDecl setter = nf.MethodDecl(c, get_set_flags, fd_type, 
                                        c.fieldSetterName(fd_name), 
                                        setter_formals, setter_stmts);   
      members.add(setter);
      ct.addMethod(setter.methodInstance());
    }
  }

  private static FieldDecl STATIC_LOCK_HOLDER;

  private FieldDecl STATIC_LOCK_HOLDER(AtomJavaContext c,
                                       AtomJavaNodeFactory nf,
                                       AtomJavaTypeSystem ts) {
    if(STATIC_LOCK_HOLDER != null)
      return STATIC_LOCK_HOLDER;

    STATIC_LOCK_HOLDER = 
      nf.FieldDecl(c, Flags.FINAL.set(Flags.PRIVATE.set(Flags.STATIC)), 
                   nf.CanonicalTypeNode(c, ts.AObject()), "staticLockHolder", 
                   nf.New(c,null,nf.CanonicalTypeNode(c,ts.AObject()),
                          Collections.EMPTY_LIST, null));
    return STATIC_LOCK_HOLDER;
  }
}

