package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.util.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

// changes all typed expressions to have types that reflect 
// the current types being used by the typesystem
public class AtomJavaBuilder extends AtomJavaContextVisitor {
  
  /* members/interfaces added to make this class atomic safe */
  List<FieldDecl> field_decls;
  List<MethodDecl> method_decls;
  List<ConstructorDecl> constructor_decls;

  /* the outer AtomJavaBuilder for nested ClassDecl's */
  AtomJavaBuilder outer;

  public AtomJavaBuilder(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  private AtomJavaBuilder enterClassDecl(ClassDecl cd) {
    AtomJavaBuilder inner = (AtomJavaBuilder) super.copy();
    inner.field_decls = new LinkedList<FieldDecl>();
    inner.method_decls = new LinkedList<MethodDecl>();
    inner.constructor_decls = new LinkedList<ConstructorDecl>();
    inner.outer = this;
    return inner;
  }

  private ClassDecl leaveClassDecl(ClassDecl cd) {
    ClassBody classBody = cd.body();
    ParsedClassType ct = cd.type();

    for(Iterator<FieldDecl> i = field_decls.iterator(); i.hasNext();) {
      FieldDecl fd = i.next();
      ct.addField(fd.fieldInstance());
      classBody = classBody.addMember(fd);
    }

    for(Iterator<MethodDecl> i = method_decls.iterator(); i.hasNext();) {
      MethodDecl md = i.next();
      ct.addMethod(md.methodInstance());
      classBody = classBody.addMember(md);
    }

    for(Iterator<ConstructorDecl> i = constructor_decls.iterator(); i.hasNext();) {
      ConstructorDecl consd = i.next();
      ct.addConstructor(consd.constructorInstance());
      classBody = classBody.addMember(consd);
    }

    /* declare the type update */
    typeSystem().addTypeUpdate(cd.type(), ct);

    /* set the type of ClassDecl and its new body*/
    return cd.type(ct).body(classBody);
  }

  public void addClassChange(FieldDecl fd) { field_decls.add(fd); }
  public void addClassChange(MethodDecl md) { method_decls.add(md); }
  public void addClassChange(ConstructorDecl cd) { constructor_decls.add(cd); }

  public NodeVisitor enter(Node parent, Node n) {
    return bypassChildren(n);
  }

  /**
   * Implement a psuedo preorder traversal. Provided the node is one which
   * can have effects from this visitor then we will call build upon it.
   * However, we initially ignore all children of nodes. We eventually need to
   * look at all children, however, in case embedded within is an AtomJavaBuild
   * node.
   *
   * @param old the <code>old</code> node and the <code>n</code> node should
   *        be the same node (pointer equality). Since we skip <code>old</code>
   *        node's children it should not change.
   * @param n see <code>old</code>
   * @param v this visitor is really a dummy visitor, it simply has been used
   *        to avoid recuring through the tree in a postorder fashion via 
   *        bypassing the children of <code>n</code>
   */
  public Node leave(Node parent, Node old, Node n, NodeVisitor v) {

    Node n_ = n;
    NodeVisitor v_ = super.enter(parent, n_);

    if(n instanceof ClassDecl)
      v_ = ((AtomJavaBuilder)v_).enterClassDecl((ClassDecl)n);

    /* all AtomJavaBuild nodes will be built in a preorder fashion */
    if(n instanceof AtomJavaBuild) {
      v_ = ((AtomJavaBuild)n).buildEnter((AtomJavaBuilder)v_);
      n_ = ((AtomJavaBuild)n).build(parent, (AtomJavaBuilder)v_);
    }
    
    n_ = n_.visitChildren(v_);

    if(n_ instanceof ClassDecl)
      n_ = ((AtomJavaBuilder)v_).leaveClassDecl((ClassDecl)n_);

    return super.leave(parent, n, n_, v_);
  }
}
