package polyglot.ext.atomjava.visit;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.types.*;
import polyglot.frontend.*;
import java.util.Stack;

/**
 *
 * A node visitor that updates loops, breaks, and continues
 * such that breaks and continues inside of atomic blocks
 * affect the correct loop (rather then the loop that atomic gets
 * translated into). 
 *
 */
public class BranchLabeler extends AJContextVisitor {

  // a stack which represents the unlabeled targets in the AST
  private Stack targets = new Stack();

  // the count of lables (for getting fresh labels)
  private static int labels;

  public BranchLabeler(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  private boolean unlabeledTarget(Node parent, Node n) {
    return !(parent instanceof Labeled) &&
      (n instanceof Switch ||
       n instanceof While ||
       n instanceof Do ||
       n instanceof For);
  }

  public NodeVisitor enter(Node parent, Node n) {
    
    if(unlabeledTarget(parent, n))
      if(!context().inAtomicBlock())
	targets.push(OUTSIDE);
      else
	targets.push(INSIDE);

    return super.enter(parent, n);
  }

  public Node leave(Node parent, Node old, Node n, NodeVisitor v) {

    Node _n = super.leave(parent, old, n, v);

    if(_n instanceof Branch) {
      Branch b = (Branch) _n;

      if(!unlabeledBranch(b) || !innermostTargetOutsideAtomicBlock())
	return _n;
	
      String label = (String) targets.peek();

      if(label == EMPTY) {
	  targets.pop();
	  label = "__aj_tgt_" + labels++  + "__";
	  targets.push(label);
      }

      return nodeFactory().Branch(context(), ((Branch)_n).kind(), label);
    }    

    if(unlabeledTarget(parent, _n))
      if(targets.peek() == INSIDE || targets.peek() == EMPTY)
	targets.pop();
      else	
	return 
	  nodeFactory().Labeled(context(), (String) targets.pop(), (Stmt) _n);

    return _n;
  }

  private boolean unlabeledBranch(Branch b) {
    return (b.label() == null || b.label().equals(""));
  }


  private boolean innermostTargetOutsideAtomicBlock() {
    return (targets.peek() != INSIDE);
  }

  private static final String INSIDE, OUTSIDE, EMPTY;
  static { INSIDE = "inside"; OUTSIDE = EMPTY = "outside|empty"; }
}
