package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.util.AJS;
import polyglot.visit.*;
import polyglot.frontend.*;
import java.util.*;


// mangle method calls as appropriate
   
public class CallMangler extends AtomJavaContextVisitor {

  public CallMangler(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  private static String[] threadMethodsToMangle = {
    "setName", "getName", "setPriority", "getPriority", "setDaemon", "isDaemon",
    "getThreadGroup", "getContextClassLoader", "setContextClassLoader",
    "stop", "suspend", "resume", "join" };

  private static String[] objectMethodsToMangle = { "wait" };

  public Node leave(Node parent, Node old, Node n, NodeVisitor v) {

    if(!(n instanceof Call))
      return n;

    AtomJavaContext c = ((AtomJavaContextVisitor)v).context();
    AtomJavaTypeSystem ts = typeSystem();
    AtomJavaNodeFactory nf = nodeFactory();
    
    Call call = (Call)n;
    
    String name = call.name();

    if(ts.isSubtype(call.target().type(), ts.AThread()))
      for(int i=0; i < threadMethodsToMangle.length; ++i)
        if(name.equals(threadMethodsToMangle[i])) {
          call = call.name("__aj_mangle_" + threadMethodsToMangle[i]);
          MethodInstance mi = call.methodInstance();
          mi = mi.name(call.name());
          call = call.methodInstance(mi);
          return call;
        }

    if(ts.isSubtype(call.target().type(), ts.AObject()))
      for(int i=0; i < objectMethodsToMangle.length; ++i)
        if(name.equals(objectMethodsToMangle[i])) {
          call = call.name("__aj_mangle_" + objectMethodsToMangle[i]);
          MethodInstance mi = call.methodInstance();
          mi = mi.name(call.name());
          call = call.methodInstance(mi);
          return call;
        }

    if(call.target().type().isArray())
      if(name.equals("clone")) {
        call = nf.Call(Position.COMPILER_GENERATED, 
                       nf.CanonicalTypeNode(c, ts.AArray()),
                       "clone", AJS.List(call.target()));
        try {
          ReferenceType ref = call.isTargetImplicit()
            ? c.currentClass() : (ReferenceType) call.target().type();
          
          MethodInstance mi = 
            c.typeSystem().findMethod(ref, call.name(), 
                                      AJS.Expr_to_Type(call.arguments()), 
                                      c.currentClass());

          call = (Call) call.type(mi.returnType());
          return call.methodInstance(mi);
        } catch(SemanticException e) { 
          throw new InternalCompilerError
            ("MethodInstance not found!: " + call, e); 
        }
      }
        
    return call;
  }

}
