package polyglot.ext.atomjava.visit;

import polyglot.ast.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.ext.atomjava.util.*;
import polyglot.ext.jl.types.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

// Note: currently this is it's own pass (not mixed in with ClassBody)
// because it will run before we add STATIC_LOCK_HOLDER. furthermore
// it requires no Type updates so it can sit anywhere.
public class ClassInitRewriter extends AtomJavaContextVisitor {

  public ClassInitRewriter(Job j, TypeSystem ts, NodeFactory nf) {
    super(j, ts, nf);
  }

  public Node leave(Node parent, Node old, Node n, NodeVisitor v) {
    
    n = super.leave(parent, old, n, v);

    if(!(n instanceof ClassBody))
      return n;
    ClassBody cb = (ClassBody)n;

    AtomJavaContext c = ((AtomJavaContextVisitor)v).context();
    AtomJavaTypeSystem ts = typeSystem();
    AtomJavaNodeFactory nf = nodeFactory();

    List members = cb.members();
    
    boolean need_clinit = false;
    for(Iterator i = members.iterator(); i.hasNext() && !need_clinit;) {
      ClassMember cm = (ClassMember)i.next();
      if(cm instanceof FieldDecl) {
        FieldDecl fd = (FieldDecl)cm;
        if(!fd.flags().isStatic())
          continue;
        Expr init = fd.init();
        if(init == null || init instanceof Lit) // ignore constants
          continue;
        need_clinit = true;
      } else if(cm instanceof Initializer) {
        Initializer init = (Initializer)cm;
        if(!init.flags().isStatic())
          continue;
        need_clinit = true;
      }
    }
    
    if(!need_clinit)
      return n;
    
    Call currentThread = nf.currentThread(c);
    
    Call en_clinit = nf.Call(c, currentThread, "enter_clinit", 
                             Collections.EMPTY_LIST);
    Call ex_clinit = nf.Call(c, currentThread, "exit_clinit", 
                             Collections.EMPTY_LIST);
    
    ClassType ct = c.currentClass();
    
    Initializer enter = 
      nf.Initializer(c, Flags.STATIC, 
                     nf.Block(c, AJS.List(nf.Eval(c, en_clinit))), ct);
    Initializer exit = 
      nf.Initializer(c, Flags.STATIC,
                     nf.Block(c, AJS.List(nf.Eval(c, ex_clinit))), ct);
    
    members = new LinkedList(members);
    members.add(0, enter);
    members.add(exit);
    
    return cb.members(members);        
  }
}
