package polyglot.ext.atomjava.visit;

import polyglot.frontend.Job;
import polyglot.types.TypeSystem;
import polyglot.ast.NodeFactory;
import polyglot.visit.NodeVisitor;
import polyglot.ast.*;
import polyglot.ext.atomjava.ast.*;
import polyglot.types.Flags;
import polyglot.types.ClassType;
import java.util.List;
import java.util.LinkedList;

// This is a little prettification/performance post-pass that
// hoists calls to AThread.currentThread().  It is currently
// _very_ conservative; only doing so if the first statement
// calls currentThread (currently true for all non-leaf methods) 
// (Note this can still increase register pressure.)

// DO ME: use a real flow-analysis to better hoist redundancies

// DO ME: use context more appropriately than this boxing stuff?
public class CurrentThreadSharer extends AtomJavaContextVisitor {
  boolean   do_replacement[];    // DO ME: no reason to visit if false
  LocalDecl me_decl[];           // variable to hold currentThread()
  int       method_decl_depth[]; // punt on nested methods
  
  public CurrentThreadSharer(Job job, TypeSystem ts, NodeFactory nf) {
    super(job,ts,nf);
    do_replacement    = new boolean[1];
    method_decl_depth = new int[1];
    me_decl           = new LocalDecl[1];
    do_replacement[0]    = false;
    method_decl_depth[0] = 0;
  }

  private boolean receiverIsClass(Receiver r, ClassType c) {
    return r instanceof TypeNode &&
      typeSystem().equals(((TypeNode)r).type(), c);
  }
  private boolean isCurrentThread(Node n) {
    if(!(n instanceof Call)) return false;
    Call c = (Call)n;
    return receiverIsClass(c.target(), typeSystem().AThread()) &&
      c.name().equals("currentThread") && 
      c.arguments().size() == 0;
  }
  private boolean isObjectGetter(Node n) {
    if(!(n instanceof Call)) return false;
    Call c = (Call)n;
    return c.arguments().size() == 1 
      && c.name().startsWith("__aj_get_");//DO ME: reading strings == ugly hack!

  }
  private boolean isArrayGetter(Node n) {
    if(!(n instanceof Call)) return false;
    Call c = (Call)n;
    return receiverIsClass(c.target(), typeSystem().AArray()) &&
      c.arguments().size() == 2 &&
      c.name().startsWith("get_");//DO ME: reading strings == ugly hack!
  }

  public NodeVisitor enterCall(Node n) {
    if(!(n instanceof CodeDecl)) return this;
    ++method_decl_depth[0];
    if(method_decl_depth[0] > 1) return this; // punt on nested methods
    // DO ME: beef-up this test
    if(((CodeDecl)n).body() == null) return this; // DAN: added this, need more beefing?
    List stmts = ((CodeDecl)n).body().statements();
    if(stmts == null || stmts.size() == 0) return this;
    Stmt first = (Stmt)stmts.get(0);
    if(!(first instanceof Eval)) return this;
    Expr e = ((Eval)first).expr();
    if(e instanceof Field)
      do_replacement[0] = isCurrentThread(((Field)e).target());
    if(e instanceof Call) 
      do_replacement[0] = isCurrentThread(((Call)e).target());
    if(do_replacement[0]) {
      AtomJavaNodeFactory nf = nodeFactory();
      me_decl[0] = nf.freshLocalDecl(context(),Flags.FINAL,"me",
                                     typeSystem().AThread(),
                                     nf.currentThread(context()));
    }
    return this;
  }

  public Node leaveCall(Node old, Node n, NodeVisitor v) {
    if(n instanceof CodeDecl) {
      --method_decl_depth[0];
      if(!do_replacement[0] || method_decl_depth[0] > 0) return n;
      do_replacement[0] = false;
      List stmts = new LinkedList(((CodeDecl)n).body().statements());
      stmts.add(0, me_decl[0]);
      return ((CodeDecl)n).body(nodeFactory().Block(context(),stmts));
    } 
    if(!do_replacement[0] || method_decl_depth[0] > 1) return n;
    if(isCurrentThread(n)) return nodeFactory().Local(context(),me_decl[0]);
    if(isObjectGetter(n) || isArrayGetter(n)) { 
      // assumes this other method already exists
      Call c    = (Call)n;
      List args = new LinkedList(c.arguments());
      args.add(0,nodeFactory().Local(context(),me_decl[0]));
      return c.arguments(args);
    }
    return n;
  }
}
